#!/usr/bin/env python3
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import input
from builtins import str
from builtins import object
import http.server
import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse
import getpass
import sys
import gi
from gi.repository.GLib import GError
from argparse import ArgumentParser
from functools import partial

from pcs_proxy.common import settings, debug_log
from pcs_proxy.server_common import CollaboraServiceProxyLogic


##
## Authentication toolkit below
##
## This class handles authentication and stores credentials either in the keyring
## if it is available, or in memory otherwise.
class Authenticator(object):
    # Try to use LibSecret if available
    try:
        gi.require_version('Secret', '1')
        from gi.repository import Secret
        # This is defined by libsecret for migration from gnome-keyring, but it
        # isn't exported by gobject-introspection, so we redefine it here.
        SECRET_SCHEMA_COMPAT_NETWORK = Secret.Schema.new(
            "org.gnome.keyring.NetworkPassword",
            Secret.SchemaFlags.NONE,
            {
                "user": Secret.SchemaAttributeType.STRING,
                "domain": Secret.SchemaAttributeType.STRING,
                "object": Secret.SchemaAttributeType.STRING,
                "protocol": Secret.SchemaAttributeType.STRING,
                "port": Secret.SchemaAttributeType.INTEGER,
                "server": Secret.SchemaAttributeType.STRING,
                "authtype": Secret.SchemaAttributeType.STRING,
            }
        )
    except ImportError:
        print("LibSecret not found. You will not be able to use the password keyring.")
        Secret = None

    def __init__(self):
        self.keyring_ok = False
        self.keyring_data = None
        # in-RAM stored username
        self.stored_u = ""
        # in-RAM stored password
        self.stored_p = ""

    def find_in_keyring(self, uri):
        """Attempts to load a username and password from the keyring, if the
        keyring is available"""
        if self.Secret is None:
            return

        username = None
        password = None

        try:
            attrs = {
                "domain": uri,
                "server": uri,
                "protocol": "https",
            }
            service = self.Secret.Service.get_sync(0, None)
            # This doesn't give us the password; only details from the schema
            results = service.search_sync(self.SECRET_SCHEMA_COMPAT_NETWORK,
                                          attrs, 0, None)
            if results:
                username = results[0].get_attributes()['user']
                # This gives us only the password
                password = self.Secret.password_lookup_sync(
                    self.SECRET_SCHEMA_COMPAT_NETWORK, attrs, None)
                # Data is already saved in keyring
                self.keyring_data = (username, password)
        except GError as e:
            # Couldn't contact daemon, or other errors
            print(str(e))

        # Keyring can be used
        self.keyring_ok = True

    def save_to_keyring(self, uri, username, password):
        try:
            attrs = {
                "user": username,
                "domain": uri,
                "server": uri,
                # BUG: Passing 'None' for a string causes a segfault
                # https://bugzilla.gnome.org/show_bug.cgi?id=685394
                "object": "",
                "protocol": "https",
            }
            self.Secret.password_store_sync(
                self.SECRET_SCHEMA_COMPAT_NETWORK, attrs,
                self.Secret.COLLECTION_DEFAULT, "PCS-Proxy password",
                password, None)
        except GError as e:
            # Couldn't contact daemon, or other errors
            print(str(e))

    # Store the baseline credentials for accessing downstream resources.
    #
    # u_app  str()  the username to use for downstream access
    # p_app  str()  the password to use for downstream access
    def set_baseline_credentials(self, u_app, p_app):
        self.stored_u = u_app
        self.stored_p = p_app

    # Prompt a user for input, and keep prompting until they provide some
    #
    # Returns the string value of their input (not stripped)
    #
    # prompt    str()    what to prompt the user with
    # password  boolean  If True, disable TTY echoing of user's typed characters
    def get_nonblank_user_input(self, prompt, password=False):
        rv_v = ""
        first = True

        while not rv_v.strip():

            if not first:
                print("Invalid, please try again!")
            else:
                first = False

            if not password:
                rv_v = input(prompt)
            else:
                rv_v = getpass.getpass(prompt)

        return rv_v

    # Get typed username and password from the user
    #
    # Returns a two-item list() with the following format:
    # [str(username), str(password)]
    def gather_username_and_password(self):
        l_u = self.get_nonblank_user_input('Username: ')
        l_p = getpass.getpass("Password for %s: " % (l_u))

        return [l_u, l_p]

    # Ask the user to re-confirm their entered credentials after a period of
    # inactivity
    #
    # Returns True if the entered credentials match those stored previously in
    #   RAM
    # Returns False otherwise
    def reconfirm_credentials(self):
        (l_u, l_p) = self.gather_username_and_password()

        return l_u == self.stored_u and l_p == self.stored_p

class CollaboraServiceProxyHandler(http.server.BaseHTTPRequestHandler):

    def __init__(self, proxy_logic, settings, *args, **kwargs):
        self.settings = settings
        self.proxy_logic = proxy_logic
        # Setup for handling outbound connections
        self.chronophage_passwords = \
                urllib.request.HTTPPasswordMgrWithDefaultRealm()
        self.chronophage_opener = \
                urllib.request.build_opener(
                    urllib.request.HTTPBasicAuthHandler(self.chronophage_passwords))
        super().__init__(*args, **kwargs)

    def handle_request(self, requests, response):
        on_index = 0

        for req_item in requests:
            on_index = on_index + 1
            real_req = None
            requesting = True
            self.settings['lastFail'] = ""
            self.settings['lastInfo'] = ""

            while requesting:
                rs = (self.proxy_logic.request_URL() + self.path)

                if not rs.endswith(req_item.lstrip("/?")):
                    rs = rs + "&" + req_item.lstrip("/?")

                response = self.settings['ProxyLogicProvider'].handle_request(rs)
                if response is not None:
                    requesting = False
                    result = response
                    continue

                method = self.settings['ProxyLogicProvider'].request_method()
                if method == "POST":
                    debug_log("\nRequest of POST type going out to:\n  " + rs + "\n\n")

                    try:
                        real_req = urllib.request.Request(rs, data=urllib.parse.urlencode(self.settings['ProxyLogicProvider'].parse_request_str(rs, True)))
                    except TypeError:
                        # This allows us to deal with parse_request_str failures below...
                        pass
                elif method == "GET":
                    debug_log("\nRequest of GET type going out to:\n  " + rs + "\n\n")
                    real_req = urllib.request.Request(rs)
                else:
                    print("\nRequest type " + str(method) + " is not supported\n\n")
                    return

                requesting = False

                if self.settings['lastFail'] == "":
                    request = self.chronophage_opener.open(real_req)
                    result = request.read()
                    requesting = self.settings['ProxyLogicProvider'].should_retry_request_after_reauthentication(result)

                    if self.settings['do_record'] and not requesting:
                        with open('transfer-log', 'a') as f:
                            output = ['>>>>']
                            output.append(real_req.get_full_url())
                            output.append(real_req.get_data())
                            output.append('<<<<')
                            output.append(u'{}\n'.format(str(result)))
                            f.write(u'\n'.join(output).encode('utf-8'))
                else:
                    result = self.settings['lastFail']

            if on_index == 1:
                response = 200

                try:
                    response = request.getcode()
                except UnboundLocalError:
                    # Handle special error cases from parse_request_str
                    pass

                self.send_response(response)
                self.end_headers()

            if on_index == 1:
                if on_index != len(requests):
                    self.wfile.write(self.settings['ProxyLogicProvider'].result_filter(result, rs, True, False) + b"\n")
                else:
                    self.wfile.write(self.settings['ProxyLogicProvider'].result_filter(result, rs, True, True) + b"\n")
                debug_log("\nRESULT PART (added to parsed message):\n")
                debug_log(result)
            else:
                if on_index == len(requests):
                    self.wfile.write(self.settings['ProxyLogicProvider'].result_filter(result, rs, False, True) + b"\n")
                    debug_log("\nRESULT PART (added to parsed message):\n")
                    debug_log(result)
                else:
                    self.wfile.write(self.settings['ProxyLogicProvider'].result_filter(result, rs, False, False) + b"\n")
                    debug_log("\nRESULT (the parsed message):\n")
                    debug_log(result)

    # Handle HTTP server GET requests
    def do_GET(self):
        settings['ProxyLogicProvider'].require_authentication()
        self.chronophage_passwords.add_password(None, self.proxy_logic.request_URL(), self.settings['credo_u'], self.settings['credo_p'])
        self.settings['ProxyLogicProvider'].session_ensure()

        self.protocol_version = 'HTTP/1.1'
        requests = self.settings['ProxyLogicProvider'].build_outgoing_requests(self)

        if requests is None:
            prevalidation_result = 'Incompatible PCS: you may need to updated the PCS macros'
        else:
            prevalidation_result = 'OK'

        response = 200

        if prevalidation_result != "OK":
            self.send_response(response)
            self.end_headers()
            self.wfile.write(self.settings['ProxyLogicProvider'].result_filter(prevalidation_result, self.proxy_logic.request_URL(), True, True))
        else:
            self.handle_request(requests, response)


def main():
    # Command line argument interface
    help_msg = sys.argv[0] + ": smart proxy server for Collabora internal tools. "

    epilog = """
    Each module has a built-in default server address to use, but you may
    pass a different SERVER (hostname) to use if you are testing changes.
    """

    cli_parser = ArgumentParser(description=help_msg, epilog=epilog)
    cli_parser.add_argument('-d', '--debug', action='store_true',
                            help="Enable output of lots of useful debugging data")
    cli_parser.add_argument('-r', '--record', action='store_true',
                            help="Record requests and responses in a 'transfer-log' file")
    cli_parser.add_argument('-u', '--username', type=str,
                            help="Username (useful for testing)")
    cli_parser.add_argument('-p', '--password', type=str,
                            help="Password (useful for testing)")
    cli_parser.add_argument('module', choices=['crm', 'chronophage'])
    cli_parser.add_argument('server', type=str, nargs='?', help="Alternate server")
    cli_parser.add_argument('api_server', type=str, nargs='?', help="Alternate API server")
    args = cli_parser.parse_args()

    CollaboraServiceProxyLogic = None
    # bring in the appropriate proxying module ("PCS to Chronophage" is default)
    if args.module == 'crm':
        from pcs_proxy.server_crm import CollaboraServiceProxyLogicCRM
        CollaboraServiceProxyLogic = CollaboraServiceProxyLogicCRM
    else:
        from pcs_proxy.server_chronophage import CollaboraServiceProxyLogicChronophage
        CollaboraServiceProxyLogic = CollaboraServiceProxyLogicChronophage

    settings['do_debug'] = args.debug
    settings['do_record'] = args.record

    # get username and password (supports CLI passage of credos for
    # ease in testing)
    credential_username = args.username
    credential_password = args.password

    # make proxy logic provider available globally
    settings['ProxyLogicProvider'] = CollaboraServiceProxyLogic()

    if args.server:
        if args.module == 'crm':
            if not args.api_server:
                print("Please specify an API server for the alternate CRM server")
                exit(1)
            else:
                settings['ProxyLogicProvider'].service_api = args.api_server

        print("Using alternate server: " + args.server)
        if args.api_server:
            print("Using API server: " + args.api_server)
        settings['ProxyLogicProvider'].service_hostname = args.server

    if settings['do_debug']:
        print("Debugging mode enabled!")

    settings['Authenticator'] = Authenticator()

    # Skip asking for credentials if they were passed from the command line.
    if not (credential_username and credential_password):
        # Grab credentials, or get them from the keyring if we can
        settings['Authenticator'].find_in_keyring(settings['ProxyLogicProvider'].service_hostname)
        if settings['Authenticator'].keyring_data:
            debug_log("KEYRING: There are saved credentials in your login keyring, those have been used!")
            credential_username, credential_password = settings['Authenticator'].keyring_data
        else:
            credential_username, credential_password = settings['Authenticator'].gather_username_and_password()

    # make credos available to the downline modules
    settings['credo_u'] = credential_username
    settings['credo_p'] = credential_password

    settings['lastFail'] = ""
    settings['lastInfo'] = ""

    # setup to start handling connections
    # Listen on 127.0.0.1 specfically to avoid isseus with python not dealing with
    # IPV6 properly
    proxy_handler = partial(CollaboraServiceProxyHandler, CollaboraServiceProxyLogic, settings)
    pcs_server = http.server.HTTPServer(('127.0.0.1', CollaboraServiceProxyLogic.port_TCP()), proxy_handler)

    # perform first-pass authorization if the proxy logic provider supports that
    settings['ProxyLogicProvider'].require_authentication()

    # A KeyError exception is thrown if authentication fails, catch this exception
    # and exit with a more friendly message.
    try:
        # peform first-time SID negotiation for proxy logic providers that need it
        settings['ProxyLogicProvider'].session_ensure()
    except KeyError as k:
        print("Session authentication failed:", k)
        exit(1)

    # setup baseline credentials if the authentication logic provider supports it
    settings['Authenticator'].set_baseline_credentials(settings['credo_u'], settings['credo_p'])

    # Don't save credentials to keyring if they were passed from the CLI.
    if not (args.username and args.password):
        # Only save password if keyring is available and after ensuring the session
        # credentials.
        if settings['Authenticator'].keyring_ok:
            settings['Authenticator'].save_to_keyring(
                settings['ProxyLogicProvider'].service_hostname,
                settings['credo_u'],
                settings['credo_p'])

    # start handling connections
    print("Ready to serve requests to " + CollaboraServiceProxyLogic.request_URL())
    print("Listening on port " + str(CollaboraServiceProxyLogic.port_TCP()) + "...")
    pcs_server.serve_forever()
